/*
 * Copyright (c) 2003-2005 Erez Zadok
 * Copyright (c) 2003-2005 David P. Quigley
 * Copyright (c) 2003-2005 Charles P. Wright
 * Copyright (c) 2003-2005 Stony Brook University
 * Copyright (c) 2003-2005 The Research Foundation of State University of New York
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package.
 *
 * This Copyright notice must be kept intact and distributed with all sources.
 */
/*
 *  $Id: unionimap.c,v 1.4 2005/06/15 17:34:23 dquigley Exp $
 */

#include "unionimap.h"

/**
 *	print_usage()
 *	Function used to print out usage information 
 *	when an error is encountered
 */
void print_usage()
{
	fprintf(stderr, "unionimap version: $Id: unionimap.c,v 1.4 2005/06/15 17:34:23 dquigley Exp $\n");
	fprintf(stderr, "Distributed with Unionfs "UNIONFS_VERSION"\n");
	fprintf(stderr, "\n");
		fprintf(stderr,"Usage: unionfsimap [-f] [-r ARG:] filename path\n");
		fprintf(stderr," flag -f: create forward map with name filename\n");
		fprintf(stderr," flag -r ARG: create reverse map with name filename and path and add it to forwardmap ARG\n");
		fprintf(stderr," flag -h: prints this message then quits\n");
		fprintf(stderr," please note that -f -r and -a are mutually exclusive\n");
}

/**
 * verify_forwardmap(char *forwardmap)
 * verifies that the forward map is valid.
 *
 */
int verify_forwardmap(char *forwardmap)
{
		int bytesread = 0,err = 0;
		FILE *fwrdmap = NULL;
		struct fmaphdr header;

		fwrdmap = fopen(forwardmap,"r");
		if(!fwrdmap){
			err =  -EINVAL;
			goto out;
		}
		bytesread = fread((void*)&header,sizeof(struct fmaphdr),1,fwrdmap);
		if(bytesread < sizeof(struct fmaphdr)){
			err =  -EINVAL;
			goto out;
		}
		if(header.magic != FORWARDMAP_MAGIC || header.version != FORWARDMAP_VERSION){
			err = -EINVAL;
			goto out;
		}
out:
		if(fwrdmap){
			fclose(fwrdmap);
		}
		return err;
}

/**
* create_forwardmap(char *filename)
* verifies that the forward map is valid.
*
*/
int create_forwardmap(char *filename)
{
		int byteswritten = 0,err = 0;
		FILE *fwrdmap = NULL;
		struct fmaphdr header;
		void *table = NULL;
		uuid_t uuid;
		
		fwrdmap = fopen(filename,"r");
		if(!fwrdmap){
			err =  -EINVAL;
			goto out;
		}
		header.version = FORWARDMAP_VERSION;
		header.magic = FORWARDMAP_MAGIC;
		header.usedbranches = 0;
		uuid_generate(uuid);
		memcpy(&header.uuid,&uuid,UUID_LEN);
		byteswritten = fwrite((void *)&header,sizeof(struct fmaphdr),1,fwrdmap);
		if(byteswritten < sizeof(struct fmaphdr)){
				unlink(filename);
				err = -EIO;
				perror("Write failed: ");
				goto out;
		}	
		table = malloc(sizeof(struct bmapent[256]));
		if(!table){
				err = -ENOMEM;
				goto out;
		}
		memset(table,0,sizeof(struct bmapent[256]));
		byteswritten = fwrite(table,sizeof(struct bmapent[256]),1,fwrdmap);
		if(byteswritten < sizeof(struct bmapent[256])){
				unlink(filename);
				err = -EIO;
				perror("Write failed: ");
				goto out;
		}
out:
		if(fwrdmap){
				fclose(fwrdmap);
		}
		if(table){
				free(table);
		}

		return err;
}

int create_reversemap(char *filename, char* path, char* forwardmap)
{
		return 0;
}
int print_forwardmap(FILE *file,int debug_level)
{
		int err = 0, bytesread = 0,i;
		struct fmaphdr header;
		struct bmapent btable[256];
		char *uuid_unparsed;
		struct fmapent entry;
		
		if(!file || debug_level <= 0){
				err = -EINVAL;
				goto out;
		}	
		err = fseek(file,0L,SEEK_SET);
		if(err){
				goto out;
		}
		bytesread = fread((void *)&header,sizeof(struct fmaphdr),1,file);
		if(bytesread != sizeof(struct fmaphdr)){
				err = -EIO;
				goto out;
		}
		uuid_unparsed = (char *)malloc(37);
		if(!uuid_unparsed){
				err = -EIO;
				goto out;
		}
		memset(uuid_unparsed,0,37);
		fprintf(stdout,"Unionfs Forwardmap:\n");
		fprintf(stdout,"Magic: %d\n",header.magic);
		fprintf(stdout,"Version: %d\n",header.version);
		fprintf(stdout,"Used Branches: %d\n",header.usedbranches);
		uuid_unparse(header.uuid,uuid_unparsed);
		fprintf(stdout,"UUID: %s\n",uuid_unparsed);
		bytesread = fread((void *)&btable,sizeof(struct bmapent[256]),1,file);
		if(bytesread != sizeof(struct fmaphdr)){
				err = -EIO;
				goto out;
		}
		for(i = 0; i < header.usedbranches;i++){
				fprintf(stdout,"Branch Pair at index : %d\n",i);
				fprintf(stdout,"fsnum: %d\n",btable[i].fsnum);
				uuid_unparse(btable[i].uuid,uuid_unparsed);
				fprintf(stdout,"uuid: %s\n",uuid_unparsed);
		}		
		if(debug_level > 1){
				fprintf(stdout,"File Contents\n");
				while(fread((void *)&entry,sizeof(struct fmapent),1,file)){
						fprintf(stdout,"fsnum: %d inode: %llu\n",(uint8_t)entry.fsnum,entry.inode);
				}
		}
out:
		return err;

}
int print_reversemap(FILE* file,int debug_level)
{

		int err = 0, bytesread = 0;
		struct rmaphdr header;
		char *uuid_unparsed = NULL;
		uint64_t inode;
		
		if(!file || debug_level <= 0){
				err = -EINVAL;
				goto out;
		}	
		err = fseek(file,0L,SEEK_SET);
		if(err){
				goto out;
		}
		bytesread = fread((void *)&header,sizeof(struct fmaphdr),1,file);
		if(bytesread != sizeof(struct fmaphdr)){
				err = -EIO;
				goto out;
		}
		uuid_unparsed = (char *)malloc(37);
		if(!uuid_unparsed){
				err = -ENOMEM;
				goto out;
		}
		memset(uuid_unparsed,0,37);
		fprintf(stdout,"Unionfs Reversemap:\n");
		fprintf(stdout,"Magic: %d\n",header.magic);
		fprintf(stdout,"Version: %d\n",header.version);
		uuid_unparse(header.fwduuid,uuid_unparsed);
		fprintf(stdout,"Forward UUID: %s\n",uuid_unparsed);
		uuid_unparse(header.revuuid,uuid_unparsed);
		fprintf(stdout,"Reverse UUID: %s\n",uuid_unparsed);
		fprintf(stdout,"fsid: %04x%04x",((unsigned int *)&header.fsid)[0],((unsigned int *)&header.fsid)[1]);
		if(debug_level > 1){
				fprintf(stdout,"File Contents\n");
				while(fread((void *)&inode ,sizeof(uint64_t),1,file)){
						if(inode){
								
							fprintf(stdout,"Pos: %lu , ino: %llu\n",ftell(file),inode);
						}
				}
		}
out:
		if(uuid_unparsed){
				free(uuid_unparsed);
		}
		return err;
}
int dump_map(char *filename, int debug_level)
{
		int err = 0,bytesread = 0;
		uint32_t magic;
		FILE *file = NULL;
	
		file = fopen(filename,"r");
		if(!file){
				err = -EINVAL;
				goto out;
		}
		bytesread = fread((void *)&magic,sizeof(uint32_t),1,file);
		if(bytesread < sizeof(uint32_t)){
				err = -EINVAL;
				goto out;
		}
		if(magic == FORWARDMAP_MAGIC){
				err = print_forwardmap(file,debug_level);
		} else if(magic == REVERSEMAP_MAGIC) {
				err = print_reversemap(file,debug_level);
		}
		else{
				err = -EINVAL;
		}
out:
		if(file){
				fclose(file);
		}
				
		return err;
}

int main(int argc, char** argv)
{
		int c = 0; //Temp varliable to hold our character for the switch
		int cflag = 0,aflag = 0,dflag = 0;
		int err = 0;
		char *avalue = NULL;
		if(argc < 2)
		{
				print_usage();
				err = -EINVAL;
				goto out;
		}
		while ((c = getopt (argc, argv, "a:cd")) != -1)
		{
				switch(c)
				{
						case 'c':
								cflag = c;
								break;
						case 'a':
								aflag = 1;
								avalue = (char*)malloc(sizeof(char)*(strlen(optarg)+1));
								strcpy(avalue,optarg);
								break;
						case 'd':
								dflag++;
								break;
						case '?':
								if(isprint (optopt))
										fprintf (stderr, "Unknown option `-%c'.\n", optopt);
								else
										fprintf (stderr,"Unknown option character `\\x%x'.\n",optopt);
								print_usage();
								err = -EINVAL;
								goto out;
								break;
						default:
								print_usage();
								err = -EINVAL;
								goto out;
								break;	
				}
		}
		if(optind == argc){
			print_usage();
			err = -EINVAL;
			goto out;
		}
		if(!(cflag^aflag^dflag)){
			print_usage();
			err = -EINVAL;
			goto out;
		}
		if(cflag){
				err = create_forwardmap(argv[optind]);
				if(err){
					unlink(argv[optind]);
				}
				goto out;
		} else if(aflag){
				if(optind+1 > argc){
					print_usage();
					err = -EINVAL;
					goto out;
				}
						
				err = create_reversemap(argv[optind],argv[optind+1],avalue);
				if(err){
					unlink(argv[optind]);
					goto out;
				}
				
		} else if(dflag){
				err = dump_map(argv[optind],dflag);
				if(err){
					goto out;
				}
		}
out:
		if(avalue){
			free(avalue);
		}
		return err;
}

/*
 * vim:shiftwidth=8
 * Local variables:
 * c-basic-offset: 8
 * End:
 */
